/*
	PWM SIN
	Date:		22/10/2014
	Target:	Cyclone IV E
	Board:	DE0-Nano
	
 OBIETTIVI:
 ----------
 - Implementare una PWM a 2 Livelli che realizzi una sinusoide
 - Implementare eventualmente la comunicazione vJTAG per variare la frequenza

 NOTE IMPLEMENTATIVE:
 --------------------
 - Questo software è stato realizzato per FPGA Cyclone III e per precauzione verranno utilizzati
  gli stessi clock, quindi 25MHz e 12.5MHz. Verranno ottenuti da un PLL
 - Per avere a 1ms 2^32 occorre uno step di 171798, 69184
 - Sin e Cos sono memorizzati a 15 bit in modo che poi il segno venga gestito successivamente
 - La creazione dell'angolo viene fatta con due banchi di RAM, mentre sarebbe meglio creare
  una pipe per calcolare l'angolo del seno e del coseno e otenere i due valori.
 
 NOTA IMPORTANTISSIMA DI TEMPORIZZAZIONE:
 ----------------------------------------
 - Per risolvere il problema del passaggio per lo 0 occorre sistemare le temporizzazioni.
  Es: se clk=25 e clk_=12.5 si mette: Theta a 25M, Sin e Cos a 25M e il cambio segno a 12.5M.
  In altre parole il cambio segno deve essere più lento del resto
  Se invece è Theta e cambio segno a 25M e SinCos a 12.5 o 50M non va perchè nell'istante in
  cui theta cambia uno degli ultimi 2 bit, il sistema si aggiorna temporaneamente.
  
 RISULTATO:
 ----------
 Funziona: vengono creati seni e coseni senza errori nel passaggio per lo 0. Accade infati
 che con clock errati nel passaggio per lo 0 si passa per un attimo negativi e poi a 0;
 per esepio: 22 >> 16 >> -6 >> 0 >> -6 >> -16 >> -22 ...
*/

module main
(	input 	clk1,			// clock 50MHz
	output	[7:0] LED,		// led
	output	P,				// PWM +: PIN_P11 (GPIO_110) pin 10 GPIO1
	output	N);				// PWM -: PIN_T11 (GPIO_17) pin 15 GPIO1

// Definizione dei clock
wire	clk2;					// clock a 25MHz		50/2
wire	clk4;					// clock a 12.5MHz	50/4

wire	sync;					// segnale di sync per evitare le commutazioni multiple

wire signed [14:0] sin;	// seno a 15 bit in uscita dalla RAM
wire [13:0]	angle;		// angolo ridotto a 0-pi/2 (è un'altra triangolare di fatto)
reg signed [15:0] duty;	// Duty-Cycle
reg	[31:0] theta = 0;	// angolo effettivo 0-2pi da cui otterremo "angle"

wire	p;						// PWM positiva senza dead time
reg	slope;
reg	[31:0] carrier;	// triangolare per realizzare la PWM
reg	[31:0] cnt;			// contatore da cui ottenere la carrier


// Creazione dei due clock aggiuntivi
pll pll0
(	.inclk0 ( clk1 ),
	.c0 ( clk2 ),
	.c1 ( clk4 ));


// Generazione dell'angolo: più il clock è lento, maggiore è la risoluzione numerica
// dello step, ma minore è la risoluzione dell'angolo.
always @(posedge clk2)
begin	: AngleGeneration
	theta <= theta + 30'd8590; //step 8590 per avere 50Hz con un clock di 25MHz
end

// L'angolo è normalizzato a 14 bit ed è una triangolare
assign angle = theta[29:16] ^ {14{theta[30]}};


// Creazione del seno ridotto: è di fatto 1/4 di seno.
// Si entra con Angle e si esce con 1/4 di seno sempre positivo.
rom rom0
(	.address ( angle ),
	.clock ( clk2 ),
	.q ( sin ) );

	
// Generazione del Duty-Cycle a 16bit con segno
// --------------------------------------------
always @(negedge clk4)	// il clock deve essere più lento di quello che genera theta
begin
	if(theta[31])
		duty <= -{1'b0, sin[14:0]};
	else
		duty <= +{1'b0, sin[14:0]};
end


// Generazione della Carrier con clock a 50MHz per ottenere 2kHz di switching
// --------------------------------------------------------------------------
always @(posedge clk1)
begin	: Counter
	slope = cnt[31];						// lo metto qui con l'"=" per essere certo di passare un ciclo. Con negedge non funziona.
	cnt = cnt + 32'd171799;				// step: 171799 = 2kHz @ clk=50M;
	//cnt = cnt + 32'd858993;			// step: 858993 = 10kHz @ clk=50M;
	carrier = (((cnt ^ {32{cnt[31]}}) - {30{1'b1}}) <<< 1'd1);	// NOTA: si perde comunque il bit meno significativo!
end

assign sync = slope ^ cnt[31];


// Generazione di una gamba PWM
// ----------------------------
PWM pwm0
(	.clk(clk1),
	.syncEnable(sync),
	.mod({duty, {16{1'b0}}}),	// modulante a 32 bit signed
	.carrier(carrier),
	.P(p) );

// Applicazione del tempo morto: in questo esempio non servirebbe,
// ma così si realizza un caso più realistico.
DeadTime
(	.clk(clk1),
	.dt(10'd250),		// 250@50MHz = 5us di delay, 50 = 1us
	.p(p), .n(!p),
	.P(P), .N(N) );
	

// Assegnazione dei led
assign LED = {{6{1'b0}}, P, N};

endmodule
